LIBRARY test_layer_export_version_2
EXPORTS
    vkEnumerateInstanceLayerProperties
    vkEnumerateInstanceExtensionProperties
    vkEnumerateDeviceLayerProperties
    vkEnumerateDeviceExtensionProperties
    test_layer_GetInstanceProcAddr
    test_layer_GetDeviceProcAddr
    GetInstanceProcAddr
    GetDeviceProcAddr
    vk_layerGetPhysicalDeviceProcAddr
    vkNegotiateLoaderLayerInterfaceVersion
    test_preinst_vkEnumerateInstanceLayerProperties
    test_preinst_vkEnumerateInstanceExtensionProperties
    test_preinst_vkEnumerateInstanceVersion
    test_override_vkGetInstanceProcAddr