//
// Copyright (c) 2014-2022 The Khronos Group Inc.
// Copyright (c) 2014-2022 Valve Corporation
// Copyright (c) 2014-2022 LunarG, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: David Pinedo <david@lunarg.com>
// Author: Charles Giessen <charles@lunarg.com>
//

#include "winres.h"

// All set through CMake
#define VER_FILE_VERSION 1, 0, 1111, 2222
#define VER_FILE_DESCRIPTION_STR "Vulkan Loader - Dev Build"
#define VER_FILE_VERSION_STR "1.0.1111.2222.Dev Build"

VS_VERSION_INFO VERSIONINFO
 FILEVERSION VER_FILE_VERSION
 PRODUCTVERSION VER_FILE_VERSION
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS VS_FF_DEBUG
#else
 FILEFLAGS 0x0L
#endif

 FILEOS 0x00000L
 FILETYPE VFT_DLL
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "04090000"
        BEGIN
            VALUE "FileDescription", VER_FILE_DESCRIPTION_STR
            VALUE "FileVersion", VER_FILE_VERSION_STR
            VALUE "LegalCopyright", "Copyright (C) 2015-2022"
            VALUE "ProductName", "Vulkan Runtime"
            VALUE "ProductVersion", VER_FILE_VERSION_STR
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 0000
    END
END
